/*
Copyright (c) 1993, Clemens Wagner
 
Permission to use, copy, modify, and distribute this software and 
its documentation for any purpose is hereby granted without fee, provided
that
(i) you have alays to distribute the whole archive
(ii) the above copyright notices and this permission notice appear in
all copies of the software and related documentation, changed or new
code has to be marked
(iii) else, there are the same conditions as for InterViews

THE SOFTWARE IS PROVIDED "AS-IS" AND WITHOUT WARRANTY OF ANY KIND, 
EXPRESS, IMPLIED OR OTHERWISE, INCLUDING WITHOUT LIMITATION, ANY 
WARRANTY OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.  
 */

#ifndef GRAPHIC_GRAPHIC_H
#define GRAPHIC_GRAPHIC_H

#include <InterViews/glyph.h>

#include "classes.h"
#include "globals.h"

class Brush;
class Color;
class Command;
class Font;
class GlyphViewer;
class Handles;
class Transformer;

const int buf_size = 10;

class Graphic: public Glyph
{
	public:
		Graphic(Graphic * = nil);
		virtual ~Graphic();

		virtual void request(Requisition &) const;
		virtual void allocate(Canvas *,const Allocation &,Extension &);
		virtual void draw(Canvas *,const Allocation &) const;
		virtual void drawit(Canvas *);
		virtual void smart_drawclipped(Canvas *,Coord,Coord,Coord,Coord);
		virtual void smart_draw(Canvas *);
		virtual void drawclipped(Canvas *,Coord,Coord,Coord,Coord);
	
		virtual Glyph *clone() const;
		virtual void flush();

		virtual Transformer *transformer() const;
		virtual void transformer(Transformer *);
		void eqv_transformer(Transformer &);

		void brush(const Brush *);
		const Brush *brush() const;
		void stroke(const Color *);
		const Color *stroke() const;
		void fill(const Color *);
		const Color *fill() const;
		void font(const Font *);
		const Font *font() const;
		void closed(boolean);
		boolean closed() const;
		void curved(boolean);
		boolean curved() const;
		int ctrlpts(Coord *&, Coord *&) const;
		void ctrlpts(Coord *,Coord *,int);

		virtual Graphic *parent();
		virtual void parent(Graphic *);
		virtual Graphic *root();

		void translate(Coord,Coord);
		void scale(Coord,Coord,Coord = 0.0, Coord = 0.0);
		void rotate(float,Coord = 0.0, Coord = 0.0);
		void align(Alignment,Graphic *,Alignment);

		virtual void recompute_shape();
		void smart_getbounds(Coord &,Coord &,Coord &,Coord &);
		virtual void getbounds(Coord &,Coord &,Coord &,Coord &);
		virtual void getcenter(Coord &,Coord &);	
		virtual boolean contains(PointObj &);
		virtual boolean intersects(BoxObj &);

		virtual void undraw();
		virtual void append_(Graphic *);
		virtual void prepend_(Graphic *);
		virtual void insert_(GlyphIndex,Graphic *);
		virtual void remove_(GlyphIndex);
		virtual void replace_(GlyphIndex,Graphic *);
		virtual void change_(GlyphIndex);
	
		virtual GlyphIndex count_() const;
		virtual Graphic *component_(GlyphIndex) const;
		virtual void modified_(GlyphIndex);

		virtual Graphic *first_containing(PointObj &);
		virtual Graphic *last_containing(PointObj &);
	
		virtual Graphic *first_intersecting(BoxObj &);
		virtual Graphic *last_intersecting(BoxObj &);

		virtual Graphic *first_within(BoxObj &);
		virtual Graphic *last_within(BoxObj &);

		virtual Graphic &operator=(Graphic &);
	
		virtual void disconnect_all();
		virtual void update(GlyphViewer * = nil);

		void get_original(const Coord *&x,const Coord *&y);
		void add_point(Coord,Coord);
		void add_curve(Coord,Coord,Coord,Coord,Coord,Coord);
		void Bspline_move_to(Coord,Coord,Coord,Coord,Coord,Coord);
		void Bspline_curve_to(Coord,Coord,Coord,Coord,Coord,Coord);

		virtual void create_handles(Canvas * =nil);
		virtual void draw_handles(Canvas * =nil);
		virtual void delete_handles();

		virtual ClassId get_classid() const;
		virtual boolean is_a(ClassId) const;
	protected:
		Graphic(const Brush *,const Color *,const Color *,const Font *,boolean,boolean,int,Transformer *);

		virtual void draw_gs(Canvas *,Graphic *);
		virtual void drawclipped_gs(Canvas *,Coord,Coord,Coord,Coord,Graphic *);
		virtual void getextent_gs(Coord &,Coord &,Coord &,Coord &,float &,Graphic *);
		virtual boolean contains_gs(PointObj &,Graphic *);
		virtual boolean intersects_gs(BoxObj &,Graphic *);
		virtual void getbounds_gs(Coord &,Coord &,Coord &,Coord &,Graphic *);
	
		virtual void total_gs(Graphic &gs);
		void parentXform(Transformer &t);
		
		virtual void concat_gs(Graphic *,Graphic *,Graphic *);
		virtual void concatXform(Transformer *,Transformer *,Transformer *);
		virtual void concat(Graphic *,Graphic *,Graphic *dest);

/*   Helpers   */

		virtual boolean contains_(Graphic *,PointObj &,Graphic *);
		virtual boolean intersects_(Graphic *,BoxObj &,Graphic *);
		virtual void getbounds_(Graphic *,Coord &,Coord &,Coord &,Coord &,Graphic *);
		void total_gs_(Graphic *,Graphic &);
		void concatgs_(Graphic *,Graphic *,Graphic *,Graphic *);
		void concatXform_(Graphic *,Transformer *,Transformer *,Transformer *);
		void concat_(Graphic *,Graphic *,Graphic *,Graphic *);
		void getextent_(Graphic*, Coord&, Coord&, Coord&, Coord&, float&,Graphic*);

		void draw_(Graphic *,Canvas *,Graphic *);
		void drawclipped_(Graphic *,Canvas *,Coord,Coord,Coord,Coord,Graphic *);
		void transform_(Coord,Coord, Coord &,Coord &,Graphic *);

		void corners(Coord &,Coord &,Coord &,Coord &,const Transformer &) const;
		void invXform_gs(Coord &tx,Coord &ty,Graphic *g) const;
		void Xform_gs(Coord [],Coord [],int,Coord [],Coord [],Graphic *) const;
	protected:
		Graphic *_graphic_state;
		Handles *_handles;
		const Brush *_brush;
		const Color *_stroke;
		const Color *_fill;
		const Font *_font;
		Transformer *_t;

		boolean _closed;
		boolean _curved;
		int _ctrlpts;
		int _buf_size;
		Coord *_x;
		Coord *_y;

		Coord _xmin;
		Coord _xmax;
		Coord _ymin;
		Coord _ymax;
		Graphic *_parent;
};

#endif
