/*
Copyright (c) 1993, Clemens Wagner
 
Permission to use, copy, modify, and distribute this software and 
its documentation for any purpose is hereby granted without fee, provided
that
(i) you have alays to distribute the whole archive
(ii) the above copyright notices and this permission notice appear in
all copies of the software and related documentation, changed or new
code has to be marked
(iii) else, there are the same conditions as for InterViews

THE SOFTWARE IS PROVIDED "AS-IS" AND WITHOUT WARRANTY OF ANY KIND, 
EXPRESS, IMPLIED OR OTHERWISE, INCLUDING WITHOUT LIMITATION, ANY 
WARRANTY OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.  
 */

#include <Graphic/graphic.h>
#include <Graphic/glyphviewer.h>
#include <Graphic/handles.h>

#include <InterViews/bitmap.h>
#include <InterViews/brush.h>
#include <InterViews/canvas.h>
#include <InterViews/character.h>
#include <InterViews/color.h>
#include <InterViews/cursor.h>
#include <InterViews/display.h>
#include <InterViews/font.h>
#include <InterViews/handler.h>
#include <InterViews/hit.h>
#include <InterViews/layout.h>
#include <InterViews/tformsetter.h>
#include <InterViews/transformer.h>
#include <InterViews/window.h>
#include <IV-X11/xcanvas.h>
#include <OS/list.h>
#include <OS/math.h>
#include <OS/memory.h>
#include <OS/string.h>
#include <stdlib.h>
#include <stream.h>
#include <ctype.h>
#include <math.h>

//
//
//

Graphic::Graphic(Graphic *gr)
{
	_t=new Transformer;
	_graphic_state=nil;
	if(gr!=nil)
	{
		Coord *foo;
		
		_brush=gr->brush();
		Resource::ref(_brush);
		_stroke=gr->stroke();
		Resource::ref(_stroke);
		_fill=gr->fill();
		Resource::ref(_fill);
		_font=gr->font();
		Resource::ref(_font);
		_closed=gr->closed();
		_curved=gr->curved();
		_ctrlpts=gr->ctrlpts(foo,foo);
		_x=new Coord[buf_size];
		_y=new Coord[buf_size];
		_buf_size=buf_size;
		if(gr->transformer()!=nil)
			*_t=*gr->transformer();
	}
	else
	{
		_brush=nil;
		_stroke=nil;
		_fill=nil;
		_font=nil;
		_closed=false;
		_curved=false;
		_ctrlpts=0;
		_x=new Coord[buf_size];
		_y=new Coord[buf_size];
		_buf_size=buf_size;
	} 
	_xmin=_xmax=_ymin=_ymax=0.0;
	_parent=nil;
	_handles=nil;
}

Graphic::~Graphic()
{
	Resource::unref(_graphic_state);
	Resource::unref(_brush);
	Resource::unref(_stroke);
	Resource::unref(_fill);
	Resource::unref(_font);
	Resource::unref(_t);
	delete _x;
	delete _y;
}

void Graphic::request(Requisition &req) const
{
	if(_ctrlpts>0)
	{
		Coord left=_xmin,bottom=_ymin;
		Coord right=_xmax,top=_ymax;
		Requirement &rx=req.x_requirement();
		Requirement &ry=req.y_requirement();

		if(_t!=nil)
			corners(left,bottom,right,top,*_t);
		
		rx.natural(right-left);
		rx.stretch(0.0);
		rx.shrink(0.0);
		rx.alignment(rx.natural()!=0.0 ? -left/rx.natural() : 0.0);
		
		ry.natural(top-bottom);
		ry.stretch(0.0);
		ry.shrink(0.0);
		ry.alignment(ry.natural()!=0.0 ? -bottom/ry.natural() : 0.0);
	}
}

void Graphic::allocate(Canvas *c,const Allocation &a,Extension &ext)
{
	if(_ctrlpts>0)
	{
		Coord w=_brush==nil ? 0 : _brush->width();
		Coord x=a.x();
		Coord y=a.y();
		ext.merge_xy(c,x+_xmin-w,y+_ymin-w,x+_xmax+w,y+_ymax+w);
	}
}

void Graphic::draw(Canvas *c,const Allocation &) const
{
	if(c!=nil)
	{
		Graphic *gr=(Graphic *) this;
		CanvasDamage &cd=c->rep()->damage_;
		gr->drawclipped(c,cd.left,cd.bottom,cd.right,cd.top);
	}
}

void Graphic::drawit(Canvas *c)
{
	if(parent()==nil)
		draw_gs(c,this);
	else
	{
		Graphic gs;
		total_gs(gs);
		draw_gs(c,&gs);
	}
}

void Graphic::smart_draw(Canvas *c)
{
	if(_graphic_state!=nil)
		draw_gs(c,_graphic_state);
	else
	{
		_graphic_state=new Graphic;
		total_gs(*_graphic_state);
		Resource::ref(_graphic_state);
		draw_gs(c,_graphic_state);
	}
}

void Graphic::smart_drawclipped(Canvas *c,Coord l,Coord b,Coord r,Coord t)
{
	if(_graphic_state!=nil)
		drawclipped_gs(c,l,b,r,t,_graphic_state);
	else
		drawclipped(c,l,b,r,t);
}

void Graphic::drawclipped(Canvas *c,Coord l,Coord b,Coord r,Coord t)
{
	if(parent()==nil)
		drawclipped_gs(c,l,b,r,t,this);
	else
	{
		Graphic *gs=new Graphic;

		Resource::unref(_graphic_state);
		total_gs(*gs);
		Resource::ref(_graphic_state=gs);
		drawclipped_gs(c,l,b,r,t,gs);
	}
}

Glyph *Graphic::clone() const
{
	return nil;
}

void Graphic::flush()
{
}

Transformer *Graphic::transformer() const
{
	return _t;
}

void Graphic::transformer(Transformer *t)
{
	if(t!=nil)
		*_t=*t;
	else
		*_t=Transformer();
}

void Graphic::eqv_transformer(Transformer &total)
{
	Graphic *p=parent();
	
	if(p==nil)
		concatXform(nil,_t,&total);
	else
	{
		p->eqv_transformer(total);
		concatXform(_t,&total,&total);
	}
}

const Brush *Graphic::brush() const
{
	return _brush;
}

void Graphic::brush(const Brush *b)
{ 
	Resource::ref(b);
	Resource::unref(_brush);
	_brush=b;
}

const Color *Graphic::stroke() const
{
	return _stroke;
}

void Graphic::stroke(const Color *s)
{
	Graphic *p=this;

	if(s!=nil)
		Resource::ref(s);
	if(_stroke!=nil)
		Resource::unref(_stroke);
	_stroke=s;
	if(_graphic_state!=nil)
		while(p!=nil)
		{
			_graphic_state->stroke(s);
			p=p->_parent;
		}
}

const Color *Graphic::fill() const
{
	return _fill;
}

void Graphic::fill(const Color *f)
{
	Graphic *p=this;

	if(f!=nil)
		Resource::ref(f);
	if(_fill!=nil)
		Resource::unref(_fill);
	_fill=f;
	if(_graphic_state!=nil)
		while(p!=nil)
		{
			_graphic_state->fill(f);
			p=p->_parent;
		}
}

const Font *Graphic::font() const
{
	return _font;
}

void Graphic::font(const Font *f)
{
	if(f!=nil)
		Resource::ref(f);
	if(_font!=nil)
		Resource::unref(_font);
	_font=f;
	Resource::unref(_graphic_state);
	_graphic_state=nil;
}

void Graphic::closed(boolean c)
{
	_closed=c;
}

boolean Graphic::closed() const
{
	return _closed;
}

void Graphic::curved(boolean c)
{
	_curved=c;
}

boolean Graphic::curved() const
{
	return _curved;
}

int Graphic::ctrlpts(Coord *&x,Coord *&y) const
{
	x=_x;
	y=_y;
	return _ctrlpts;
}

void Graphic::ctrlpts(Coord *x,Coord *y,int count)
{
	delete _x;
	delete _y;
	_ctrlpts=max(count+1,buf_size);
	_x=new Coord[_ctrlpts];
	_y=new Coord[_ctrlpts];
	for(int i=0;i<count;i++)
	{
		_x[i]=x[i];
		_y[i]=y[i];
	}
}

Graphic *Graphic::parent()
{
	return _parent;
}

void Graphic::parent(Graphic *p)
{
	_parent=p;
}

Graphic *Graphic::root()
{
	Graphic *cur,*parent=this;

	do
	{
		cur=parent;
		parent=cur->parent();
	} while(parent!=nil);
	return cur;
}

void Graphic::translate(float dx,float dy)
{ 
	if(dx!=0||dy!=0)
	{
		if(_t==nil)
			_t=new Transformer;
		_t->translate(dx,dy);
		Resource::unref(_graphic_state);
		_graphic_state=nil;
	}
}

void Graphic::scale(float sx,float sy,float cx,float cy)
{
	float ncx,ncy;

	if(sx!=1.0||sy!=1.0)
	{
		Transformer parents;

		if(_t==nil)
			_t=new Transformer;
		parentXform(parents);
		parents.InvTransform(cx,cy,ncx,ncy);
	
		if(ncx!=0||ncy!=0)
		{
			_t->translate(-ncx,-ncy);
			_t->scale(sx,sy);
			_t->translate(ncx,ncy);
		}
		else
			_t->scale(sx,sy);
		Resource::unref(_graphic_state);
		_graphic_state=nil;
	}
}

void Graphic::rotate(float angle,float cx,float cy)
{
	float mag=(angle<0) ? -angle : angle;
	float ncx,ncy;

	if((mag-int(mag))!=0||int(mag)%360!=0)
	{
		Transformer parents;

		if(_t==nil)
			_t=new Transformer;
		parentXform(parents);
		parents.InvTransform(cx,cy,ncx,ncy);
	
		if(ncx!=0||ncy!=0)
		{
			_t->translate(-ncx,-ncy);
			_t->rotate(angle);
			_t->translate(ncx,ncy);
		}
		else
			_t->rotate(angle);
		Resource::unref(_graphic_state);
		_graphic_state=nil;
	}
}

void Graphic::align(Alignment falign,Graphic *moved,Alignment malign)
{
	float fx0,fy0,fx1,fy1,mx0,my0,mx1,my1,dx=0,dy=0;

	getbounds(fx0,fy0,fx1,fy1);
	moved->getbounds(mx0,my0,mx1,my1);
	
	switch(falign)
	{
		case BottomLeft:
		case CenterLeft:
		case TopLeft:
		case Left:
			dx=fx0;
		break;
		case BottomCenter:
		case Center:
		case TopCenter:
		case HorizCenter:
			dx=(fx0+fx1+1)/2;
		break;
		case BottomRight:
		case CenterRight:
		case TopRight:
		case Right:
			dx=fx1+1;
		break;
	}
	switch(falign)
	{
		case BottomLeft:
		case BottomCenter:
		case BottomRight:
		case Bottom:
			dy=fy0;
		break;
		case CenterLeft:
		case Center:
		case CenterRight:
		case VertCenter:
			dy=(fy0+fy1+1)/2;
		break;
		case TopLeft:
		case TopCenter:
		case TopRight:
		case Top:
			dy=fy1+1;
		break;
	}
	
	switch(malign)
	{
		case BottomLeft:
		case CenterLeft:
		case TopLeft:
		case Left:
			dx -= mx0;
		break;	
		case BottomCenter:
		case Center:
		case TopCenter:
		case HorizCenter:
			dx -=(mx0+mx1+1)/2;
		break;
		case BottomRight:
		case CenterRight:
		case TopRight:
		case Right:
			dx -=(mx1+1);
		break;
	}
	switch(malign)
	{
		case BottomLeft:
		case BottomCenter:
		case BottomRight:
		case Bottom:
			dy -= my0;
		break;
		case CenterLeft:
		case Center:
		case CenterRight:
		case VertCenter:
			dy -=(my0+my1+1)/2;
		break;
		case TopLeft:
		case TopCenter:
		case TopRight:
		case Top:
			dy -=(my1+1);
		break;
	}
	if(dx!=0||dy!=0)
	{
		Transformer parents;

		moved->parentXform(parents);
		parents.Invert();
		parents.Transform(0.0,0.0,fx0,fy0);
		parents.Transform(dx,dy,mx0,my0);

		moved->translate(mx0-fx0,my0-fy0);
		Resource::unref(_graphic_state);
		_graphic_state=nil;
	}
}

void Graphic::recompute_shape()
{
	int i;

	if(_ctrlpts==0)
		return;
	else
	{
		_xmin=_xmax=_x[0];
		_ymin=_ymax=_y[0];
	}
	for(i=1;i<_ctrlpts;i++)
	{
		_xmin=Math::min(_xmin,_x[i]);
		_xmax=Math::max(_xmax,_x[i]);
		_ymin=Math::min(_ymin,_y[i]);
		_ymax=Math::max(_ymax,_y[i]);
	}
	Resource::unref(_graphic_state);
	_graphic_state=nil;
}

void Graphic::smart_getbounds(float &x0,float &y0,float &x1,float &y1)
{
	if(_graphic_state!=nil)
		getbounds_gs(x0,y0,x1,y1,_graphic_state);
	else
		getbounds(x0,y0,x1,y1);
}

void Graphic::getbounds(float &x0,float &y0,float &x1,float &y1)
{
	Graphic gs;
	
	total_gs(gs);
	getbounds_gs(x0,y0,x1,y1,&gs);
}

void Graphic::getcenter(float &x,float &y)
{
	Graphic gs;
	float l,b,tol;
	
	total_gs(gs);
	getextent_gs(l,b,x,y,tol,&gs);
}	

boolean Graphic::contains(PointObj &p)
{
	if(parent()==nil)
		return contains_gs(p,this);
	else
	{
		Graphic gs;
		total_gs(gs);
		return contains_gs(p,&gs);
	}
}

boolean Graphic::intersects(BoxObj &b)
{
	if(parent()==nil)
		return intersects_gs(b,this);
	else
	{
		Graphic gs;
		total_gs(gs);
		return intersects_gs(b,&gs);
	}
}

void Graphic::undraw()
{
}

void Graphic::append_(Graphic *)
{
}

void Graphic::prepend_(Graphic *)
{
}

void Graphic::insert_(GlyphIndex,Graphic *)
{
}

void Graphic::remove_(GlyphIndex)
{
}

void Graphic::replace_(GlyphIndex,Graphic *)
{
}

void Graphic::change_(GlyphIndex)
{
}

GlyphIndex Graphic::count_() const
{
	return 0;
}

Graphic *Graphic::component_(GlyphIndex) const
{
	return nil;
}

void Graphic::modified_(GlyphIndex)
{
}


Graphic *Graphic::first_containing(PointObj &)
{
	return nil;
}

Graphic *Graphic::last_containing(PointObj &)
{
	return nil;
}

Graphic *Graphic::first_intersecting(BoxObj &)
{
	return nil;
}

Graphic *Graphic::last_intersecting(BoxObj &)
{
	return nil;
}

Graphic *Graphic::first_within(BoxObj &)
{
	return nil;
}

Graphic *Graphic::last_within(BoxObj &)
{
	return nil;
}


Graphic &Graphic::operator=(Graphic &g)
{
	brush(g.brush());
	stroke(g.stroke());
	fill(g.fill());
	font(g.font());

	if(g._t==nil)
	{
		Resource::unref(_t);
		_t=nil;

	}
	else
	{
		if(_t==nil)
			_t=new Transformer(g._t);
		else
			*_t=*g._t;
	}
	return *this;
}

void Graphic::get_original(const Coord *&x,const Coord *&y)
{
	x=_x;
	y=_y;
}

void Graphic::add_point(Coord x,Coord y)
{
	if(_ctrlpts==0)
	{
		_xmin=x;
		_xmax=x;
		_ymin=y;
		_ymax=y;
	}
	else
	{
		_xmin=Math::min(_xmin,x);
		_xmax=Math::max(_xmax,x);
		_ymin=Math::min(_ymin,y);
		_ymax=Math::max(_ymax,y);
	}
	_x[_ctrlpts]=x;
	_y[_ctrlpts]=y;
	_ctrlpts+=1;
	if(_ctrlpts>=_buf_size)
	{
		_buf_size=buf_size+_buf_size;
		Coord *x=new Coord[_buf_size];
		Coord *y=new Coord[_buf_size];
		Memory::copy(_x,x,_ctrlpts*sizeof(int));
		Memory::copy(_y,y,_ctrlpts*sizeof(int));
		delete _x;
		delete _y;
		_x=x;
		_y=y;
	}
}

void Graphic::add_curve(Coord x,Coord y,Coord x1,Coord y1,Coord x2,Coord y2)
{
	add_point(x1,y1);
	add_point(x2,y2);
	add_point(x,y);
}

void Graphic::Bspline_move_to(Coord x,Coord y,Coord x1,Coord y1,Coord x2,Coord y2)
{
	Coord p1x=(x+x+x1)/3;
	Coord p1y=(y+y+y1)/3;
	Coord p2x=(x+x+x2)/3;
	Coord p2y=(y+y+y2)/3;
	add_point((p1x+p2x)/2,(p1y+p2y)/2);
}

void Graphic::Bspline_curve_to(Coord x,Coord y,Coord x1,Coord y1,Coord x2,Coord y2)
{
	Coord p1x=(x+x+x1)/3;
	Coord p1y=(y+y+y1)/3;
	Coord p2x=(x+x+x2)/3;
	Coord p2y=(y+y+y2)/3;
	Coord p3x=(x1+x1+x)/3;
	Coord p3y=(y1+y1+y)/3;
	add_curve((p1x+p2x)/2,(p1y+p2y)/2,p3x,p3y,p1x,p1y);
}

void Graphic::disconnect_all()
{
}

void Graphic::update(GlyphViewer *)
{
	Resource::unref(_graphic_state);
	_graphic_state=nil;
}

void Graphic::create_handles(Canvas *c)
{
	Coord l,b,r,t,mx,my;

	getbounds(l,b,r,t);
	getcenter(mx,my);
	if(_handles!=nil)
	{
		Coord *x,*y;

		if(_handles->get_original(x,y)<8)
			delete_handles();
		else
		{
			x[0]=x[6]=x[7]=l;
			x[1]=x[5]=mx;
			x[2]=x[3]=x[4]=r;
			y[0]=y[1]=y[2]=b;
			y[3]=y[7]=my;
			y[4]=y[5]=y[6]=t;
		}
	}
	if(_handles==nil)
	{
		Coord x[8],y[8];

		x[0]=x[6]=x[7]=l;
		x[1]=x[5]=mx;
		x[2]=x[3]=x[4]=r;
		y[0]=y[1]=y[2]=b;
		y[3]=y[7]=my;
		y[4]=y[5]=y[6]=t;
		_handles=new Handles(x,y,8);
	}
	if(c!=nil)
	{
		Transformer *t=new Transformer;

		c->push_transform();
		c->transformer(t);
		_handles->draw(c);
		c->pop_transform();
		t->unref();
	}
}

void Graphic::draw_handles(Canvas *c)
{
	if(c!=nil)
	{
		Transformer *t=new Transformer;

		c->push_transform();
		c->transformer(t);
		create_handles();
		if(_handles->drawn())
			_handles->redraw();
		else
			_handles->draw(c);
		c->pop_transform();
		t->unref();
	}
}

void Graphic::delete_handles()
{
	if(_handles!=nil)
	{
		_handles->erase();
		delete _handles;
		_handles=nil;
	}
}

ClassId Graphic::get_classid() const
{
	return GRAPHIC;
}

boolean Graphic::is_a(ClassId id) const
{
	return id==GRAPHIC;
}

// Graphic: protected operations

Graphic::Graphic
(
	const Brush *brush,const Color *stroke,const Color *fill,const Font *font,
	boolean closed,boolean curved,int coords,Transformer *t
)
{
	_graphic_state=nil;
	_brush=brush;
	Resource::ref(_brush);
	_stroke=stroke;
	Resource::ref(_stroke);
	_fill=fill;
	Resource::ref(_fill);
	_font=font;
	Resource::ref(_font);
	_closed=closed;
	_curved=curved;
	_ctrlpts=0;
	if(coords>0)
	{
		_x=new Coord[coords];
		_y=new Coord[coords];
	}
	else
	{
		_x=nil;
		_y=nil;
	}
	_buf_size=coords;
	_t=new Transformer;
	if(t!=nil)
		*_t=*t;
	_parent=nil;
	_handles=nil;
}

void Graphic::draw_gs(Canvas *c,Graphic *gs)
{
	const Brush *brush=gs->brush();
	const Color *stroke=gs->stroke();
	const Color *fill=gs->fill();
	Transformer *tx=gs->transformer();
	int i;

	c->push_transform();
	if(tx!=nil)
		c->transform(*tx);
	c->new_path();
	c->move_to(_x[0],_y[0]);
	if(_curved)
	{
		for(i=1;i<_ctrlpts;i += 3)
			c->curve_to(_x[i+2],_y[i+2],_x[i],_y[i],_x[i+1],_y[i+1]);
	}
	else
	{
		for(i=1;i<_ctrlpts;++i)
			c->line_to(_x[i],_y[i]);
	}
	if(_closed)
		c->close_path();
	if(fill!=nil)
		c->fill(fill);
	if(brush!=nil && stroke!=nil)
		c->stroke(stroke,brush);
	/*
	if(_handles!=nil)
	{
		c->transformer(nil);
		create_handles();
		if(_handles->drawn())
			_handles->redraw();
		else
			_handles->draw(c);
	}
	*/
	c->pop_transform();
}

void Graphic::drawclipped_gs(Canvas *c,Coord l,Coord b,Coord r,Coord t,Graphic *gs)
{
	Coord ll,bb,rr,tt;
	getbounds_gs(ll,bb,rr,tt,gs);
	BoxObj thisBox(ll,bb,rr,tt);
	BoxObj clipBox(l,b,r,t);

	if(clipBox.Intersects(thisBox))
		draw_gs(c,gs);
}

void Graphic::getextent_gs(Coord &l,Coord &b,Coord &cx,Coord &cy,float &tol,Graphic *gs)
{
	tol=1.0;
	const Brush *br=gs->brush();
	if(br!=nil)
	{
		float width=float(br->Width());
		tol=(width>1) ? width : tol;
	}
	if(_ctrlpts>0)
	{
		Coord left=_xmin,bottom=_ymin;
		Coord right=_xmax,top=_ymax;
		Transformer *t=gs->transformer();

		if(t!=nil)
			corners(left,bottom,right,top,*t);
		l=left;
		b=bottom;
		cx=(left+right)/2.0;
		cy=(top+bottom)/2.0;
	}
}

boolean Graphic::contains_gs(PointObj &po,Graphic *gs)
{ 
	Coord ll,bb,rr,tt;
	PointObj pt(&po);
	getbounds_gs(ll,bb,rr,tt,gs);
	BoxObj b(ll,bb,rr,tt);;

	if(!_curved && !_fill)
	{
		if(b.Contains(po))
		{
			MultiLineObj ml(_x,_y,_ctrlpts);
			invXform_gs(pt._x,pt._y,gs);
			if(_closed)
			{
				LineObj l(_x[_ctrlpts-1],_y[_ctrlpts-1],*_x,*_y);
				return ml.Contains(pt) || l.Contains(pt);
			}
			else
				return ml.Contains(pt);
		}
		return false;

	}
	else
		if(!_curved && _fill)
		{
			if(b.Contains(pt))
			{
				FillPolygonObj fp(_x,_y,_ctrlpts);
				invXform_gs(pt._x,pt._y,gs);
				return fp.Contains(pt);
			}
			return false;
		}
		else
			if(_curved && !_fill)
			{
				if(b.Contains(pt))
				{
					MultiLineObj ml;
					if(_closed)
						ml.ClosedSplineToPolygon(_x,_y,_ctrlpts);
					else
						ml.SplineToMultiLine(_x,_y,_ctrlpts);
					invXform_gs(pt._x,pt._y,gs);
					return ml.Contains(pt);
				}
				return false;
			}
			else
			{
				if(b.Contains(pt))
				{
					FillPolygonObj fp;
					fp.ClosedSplineToPolygon(_x,_y,_ctrlpts);
					invXform_gs(pt._x,pt._y,gs);
					return fp.Contains(pt);
				}
				return false;
			}
}

boolean Graphic::intersects_gs(BoxObj &userb,Graphic *gs)
{ 
	Coord *convx,*convy;
	Coord ll,bb,rr,tt;
	getbounds_gs(ll,bb,rr,tt,gs);
	BoxObj b(ll,bb,rr,tt);;
	boolean result=false;

	if(!_curved && !_fill)
	{
		if(b.Intersects(userb))
		{
			convx=new Coord[_ctrlpts+1];
			convy=new Coord[_ctrlpts+1];
			Xform_gs(_x,_y,_ctrlpts,convx,convy,gs);
			if(_closed)
			{
				convx[_ctrlpts]=*convx;
				convy[_ctrlpts]=*convy;
				MultiLineObj ml(convx,convy,_ctrlpts+1);
				result=ml.Intersects(userb);
			}
			else
			{
				MultiLineObj ml(convx,convy,_ctrlpts);
				result=ml.Intersects(userb);
			}
			delete convx;
			delete convy;
		}
		return result;
	}
	else
	if(!_curved && _fill)
	{
		if(b.Intersects(userb))
		{
			convx=new Coord[_ctrlpts];
			convy=new Coord[_ctrlpts];
			Xform_gs(_x,_y,_ctrlpts,convx,convy,gs);
			FillPolygonObj fp(convx,convy,_ctrlpts);
			result=fp.Intersects(userb);
			delete convx;
			delete convy;
		}
		return result;	

	}
	else
		if(_curved && !_fill)
		{
			if(b.Intersects(userb))
			{
				convx=new Coord[_ctrlpts];
				convy=new Coord[_ctrlpts];
				Xform_gs(_x,_y,_ctrlpts,convx,convy,gs);
				MultiLineObj ml;
				if(_closed)
					ml.ClosedSplineToPolygon(convx,convy,_ctrlpts);
				else
					ml.SplineToMultiLine(convx,convy,_ctrlpts);
				result=ml.Intersects(userb);
				delete convx;
				delete convy;
			}
			return result;

		}
		else
		{
			if(b.Intersects(userb))
			{
				convx=new Coord[_ctrlpts];
				convy=new Coord[_ctrlpts];
				Xform_gs(_x,_y,_ctrlpts,convx,convy,gs);
				FillPolygonObj fp;
				fp.ClosedSplineToPolygon(convx,convy,_ctrlpts);
				result=fp.Intersects(userb);
				delete convx;
				delete convy;
			}
			return result;
		}
}

void Graphic::getbounds_gs(Coord &l,Coord &b,Coord &r,Coord &t,Graphic *gs)
{
	float tol;
	
	getextent_gs(l,b,r,t,tol,gs);
	r+=r-l;
	t+=t-b;
	l-=tol;
	b-=tol;
	r+=tol;
	t+=tol;
}

void Graphic::total_gs(Graphic &gs)
{
	Graphic *p=parent();

	if(p==nil)
		concat(nil,this,&gs);
	else
	{
		p->total_gs(gs);
		concat(this,&gs,&gs);
	}
}

void Graphic::parentXform(Transformer &t)
{
	Transformer identity;
	Graphic *p=parent();

	if(p==nil)
		*(&t)=*(&identity);
	else
		p->eqv_transformer(t);
}

void Graphic::concat_gs(Graphic *a,Graphic *b,Graphic *dest)
{
	const Color *stroke,*fill;
	const Font *font;
	const Brush *br;
	
	if(a==nil)
	{
		*dest=*b;
		return;
	}
	else
		if(b==nil)
		{
			*dest=*a;
			return;
		}
	if((fill=b->fill())==nil)
		fill=a->fill();
	dest->fill(fill);
	if((stroke=b->stroke())==nil)
		stroke=a->stroke();
	dest->stroke(stroke);

	if((font=b->font())==nil)
		font=a->font();
	dest->font(font);
	if((br=b->brush())==nil)
		br=a->brush();
	dest->brush(br);
}

void Graphic::concatXform(Transformer *a,Transformer *b,Transformer *dest)
{
	Transformer identity;

	if(a==nil)
		*dest=(b==nil) ? *(&identity) : *b;
	else
		if(b==nil)
			*dest=*a;
		else
		{
			Transformer tmp(a);

			tmp.Postmultiply(b);
			*dest=tmp;
		}
}

void Graphic::concat(Graphic *a,Graphic *b,Graphic *dest)
{
	Transformer *ta=(a==nil) ? nil : a->transformer();
	Transformer *tb=(b==nil) ? nil : b->transformer();
	Transformer *td=dest->transformer();
	if(td==nil)
		td=new Transformer;
	else
		Resource::ref(td);
	concatXform(ta,tb,td);
	dest->transformer(td);
	concat_gs(a,b,dest);
	Resource::unref(td);
}

boolean Graphic::contains_(Graphic *g,PointObj &p,Graphic *gs)
{
	return g->contains_gs(p,gs);
}

boolean Graphic::intersects_(Graphic *g,BoxObj &b,Graphic *gs)
{
	return g->intersects_gs(b,gs);
}

void Graphic::getbounds_(Graphic *g,Coord &l,Coord &b,Coord &r,Coord &t,Graphic *gs)
{
	g->getbounds_gs(l,b,r,t,gs);
}

void Graphic::total_gs_(Graphic *gr,Graphic &gs)
{
	gr->total_gs(gs);
}

void Graphic::concatgs_(Graphic *gr,Graphic *a,Graphic *b,Graphic *dest)
{
	gr->concat_gs(a,b,dest);
}

void Graphic::concatXform_(Graphic *gr,Transformer *a,Transformer *b,Transformer *dest)
{
	gr->concatXform(a,b,dest);
}

void Graphic::concat_(Graphic *g,Graphic *a,Graphic *b,Graphic *d)
{
	g->concat(a,b,d);
}

void Graphic::getextent_(Graphic *gr,Coord &l,Coord &b,Coord &r,Coord &t,float &tol,Graphic *gs)
{
	gr->getextent_gs(l,b,r,t,tol,gs);
}

void Graphic::draw_(Graphic *gr,Canvas *c,Graphic *gs)
{
	gr->draw_gs(c,gs);
}

void Graphic::drawclipped_(Graphic *gr,Canvas *c,Coord l,Coord b,Coord r,Coord t,Graphic *gs)
{
	gr->drawclipped_gs(c,l,b,r,t,gs);
}

void Graphic::transform_(Coord x,Coord y,Coord &tx,Coord &ty,Graphic *g)
{
	Transformer *t=(g==nil) ? transformer() : g->transformer();

	if(t!=nil)
		t->Transform(x,y,tx,ty);
	else
	{
		tx=x;
		ty=y;
	}
}

void Graphic::corners(Coord &left,Coord &bottom,Coord &right,Coord &top,const Transformer &t) const
{
	Coord x1,y1,x2,y2,x3,y3,x4,y4;
	
	t.transform(left,bottom,x1,y1);
	t.transform(left,top,x2,y2);
	t.transform(right,top,x3,y3);
	t.transform(right,bottom,x4,y4);
	left=Math::min(x1,x2,x3,x4);
	bottom=Math::min(y1,y2,y3,y4);
	right=Math::max(x1,x2,x3,x4);
	top=Math::max(y1,y2,y3,y4);
}	

void Graphic::invXform_gs(Coord &tx,Coord &ty,Graphic *g) const
{
	Transformer *t=g->transformer();
	if(t!=nil)
		t->inverse_transform(tx,ty);
}

void Graphic::Xform_gs(Coord x[],Coord y[],int n,Coord tx[],Coord ty[],Graphic *g) const
{
	Transformer *t=g->transformer();

	if(t!=nil)
	{
		register Coord *ox, *oy,*nx,*ny;
		Coord *lim;
		
		lim=&x[n];
		for(ox=x,oy=y,nx=tx,ny=ty;ox<lim;ox++,oy++,nx++,ny++)
			t->transform(*ox,*oy,*nx,*ny);
	}
	else
	{
		Memory::copy(x,tx,n*sizeof(Coord));
		Memory::copy(y,ty,n*sizeof(Coord));
	}
}
