/*
Copyright (c) 1993, Clemens Wagner
 
Permission to use, copy, modify, and distribute this software and 
its documentation for any purpose is hereby granted without fee, provided
that
(i) you have alays to distribute the whole archive
(ii) the above copyright notices and this permission notice appear in
all copies of the software and related documentation, changed or new
code has to be marked
(iii) else, there are the same conditions as for InterViews

THE SOFTWARE IS PROVIDED "AS-IS" AND WITHOUT WARRANTY OF ANY KIND, 
EXPRESS, IMPLIED OR OTHERWISE, INCLUDING WITHOUT LIMITATION, ANY 
WARRANTY OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.  
 */

#include <Graphic/connector.h>
#include <Graphic/glyphviewer.h>
#include <Graphic/handles.h>
#include <Graphic/lines.h>
#include <Graphic/link.h>

#include <InterViews/transformer.h>

Link::Link(FixedConnector *fc1,FixedConnector *fc2,const Brush *b,const Color *s,const Color *f): PolyGraphic(nil)
{
	append_(new Line(b,s,nil,0,0,1,0,nil));
	append_(_fc1=new FlexibleConnector(b,f,f,nil));
	append_(_fc2=new FlexibleConnector(b,f,f,nil));
	if(fc1!=nil)
		_fc1->connect(fc1);
	if(fc2!=nil)
		_fc2->connect(fc2);
}

Link::~Link()
{
}

Glyph *Link::clone() const
{
	Link *that=(Link *) this;
	Link *link=new Link(connector(0),connector(1),_brush,_stroke,_fill);
	Transformer *t=new Transformer(that->transformer());

	link->transformer(t);
	Resource::unref(t);
	return link;
}

FixedConnector *Link::connector(boolean last) const
{
	return (FixedConnector *) ((FlexibleConnector *) component_(last ? LINK_LAST_CONN : LINK_FIRST_CONN))->connector(0);
}

void Link::connector(FixedConnector *fix,boolean last)
{
	((FlexibleConnector *) component_(last ? LINK_LAST_CONN : LINK_FIRST_CONN))->connect(fix);
}

void Link::update(GlyphViewer *gv)
{
	Coord x,y;
	Coord x1,y1,x2,y2;
	Transformer total;
	Graphic *line=component_(0);

	transformer(nil);
	_fc1->update(gv);
	_fc2->update(gv);
	_fc1->getcenter(x1,y1);
	_fc2->getcenter(x2,y2);
	parentXform(total);
	total.invert();
	total.transform(x1,y1);
	total.transform(x2,y2);
	x2-=x1;
	y2-=y1;
	if(gv!=nil)
	{
		Transformer *tf=new Transformer(x2,y2,-y2,x2,x1,y1);
		Coord l,b,r,t;
		Coord ll,bb,rr,tt;

		line->getbounds(l,b,r,t);
		line->transformer(tf);
		line->getbounds(ll,bb,rr,tt);
		if(l!=ll || b!=bb || r!=rr || t!=tt)
		{
			gv->damage(l,b,r,t);
			gv->damage(ll,bb,rr,tt);
		}
		tf->unref();
	}
	else
	{
		Transformer *tf=new Transformer(x2,y2,-y2,x2,x1,y1);

		line->transformer(tf);
		tf->unref();
	}
	Graphic::update();
}

void Link::create_handles(Canvas *c)
{
	Coord x0,y0,x1,y1,x2,y2;
	Transformer total;
	
	get_line()->eqv_transformer(total);
	total.transform(0,0,x0,y0);
	total.transform(1,0,x2,y2);
	get_line()->getcenter(x1,y1);
	if(_handles!=nil)
	{
		Coord *x,*y;

		if(_handles->get_original(x,y)<3)
			delete_handles();
		else
		{
			x[0]=x0;
			y[0]=y0;
			x[1]=x1;
			y[1]=y1;
			x[2]=x2;
			y[2]=y2;
		}
	}
	if(_handles==nil)
	{
		Coord x[3],y[3];

		x[0]=x0;
		y[0]=y0;
		x[1]=x1;
		y[1]=y1;
		x[2]=x2;
		y[2]=y2;
		_handles=new Handles(x,y,3);
	}
	if(c!=nil)
		_handles->draw(c);
}

ClassId Link::get_classid() const
{
	return LINK_GRAPHIC;
}

boolean Link::is_a(ClassId id) const
{
	return id==LINK_GRAPHIC || PolyGraphic::is_a(id);
}

