/*
Copyright (c) 1993, Clemens Wagner
 
Permission to use, copy, modify, and distribute this software and 
its documentation for any purpose is hereby granted without fee, provided
that
(i) you have alays to distribute the whole archive
(ii) the above copyright notices and this permission notice appear in
all copies of the software and related documentation, changed or new
code has to be marked
(iii) else, there are the same conditions as for InterViews

THE SOFTWARE IS PROVIDED "AS-IS" AND WITHOUT WARRANTY OF ANY KIND, 
EXPRESS, IMPLIED OR OTHERWISE, INCLUDING WITHOUT LIMITATION, ANY 
WARRANTY OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.  
 */

#include <Graphic/slrect.h>

#include <InterViews/brush.h>
#include <InterViews/canvas.h>
#include <InterViews/color.h>
#include <InterViews/transformer.h>

//
//
//

SlidingLine::SlidingLine(Coord x1,Coord y1,Coord x2,Coord y2,Canvas *c)
{
        _fixx=x1;
        _fixy=y1;
        _flexx=x2;
        _flexy=y2;
        _c=c;
	_tf=new Transformer;
        _color=new Color(1,1,1,1,Color::Xor);
        _color->ref();
        _brush=new Brush(1);
        _brush->ref();
        draw();
}

SlidingLine::~SlidingLine()
{
	draw();
	Resource::unref(_color);
	Resource::unref(_brush);
	Resource::unref(_tf);
}

void SlidingLine::translate(Coord x,Coord y)
{
	draw();
	_flexx+=x;
	_flexy+=y;
	draw();
}

//


void SlidingLine::draw()
{
	static const Coord rw=3.0;

	_c->push_transform();
	_c->transformer(_tf);
	_c->new_path();
	_c->move_to(_fixx,_fixy);
	_c->line_to(_flexx,_flexy);
	_c->stroke(_color,_brush);
	_c->fill_rect(_fixx-rw,_fixy-rw,_fixx+rw,_fixy+rw,_color);
	_c->pop_transform();
}

//
//
//

SlidingRect::SlidingRect(Coord l,Coord b,Coord r,Coord t,Canvas *c)
{
	_l=l;
	_b=b;
	_r=r;
	_t=t;
	_tf=new Transformer;
	_c=c;
	_color=new Color(1,1,1,1,Color::Xor);
	_color->ref();
	_brush=new Brush(1);
	_brush->ref();
	draw();
}

SlidingRect::~SlidingRect()
{
	draw();
	Resource::unref(_color);
	Resource::unref(_brush);
	Resource::unref(_tf);
}

void SlidingRect::translate(Coord x,Coord y)
{
	draw();
	_tf->translate(x,y);
	draw();
}

void SlidingRect::scale(Coord sx,Coord sy)
{
	draw();
	_tf->scale(sx,sy);
	draw();
}

void SlidingRect::rotate(Coord a)
{
	draw();
	_tf->rotate(a);
	draw();
}

void SlidingRect::getrect(Coord &orgx,Coord &orgy,Coord &x,Coord &y) const
{
	_tf->transform(_l,_b,orgx,orgy);
	_tf->transform(_r,_t,x,y);
}

void SlidingRect::transformer(Transformer *t)
{
	draw();
	Resource::ref(t);
	Resource::unref(_tf);
	_tf=t;
	draw();
}

//

void SlidingRect::draw()
{
	_c->push_transform();
	_c->transformer(_tf);
	_c->new_path();
	_c->move_to(_l,_b);
	_c->line_to(_l,_t);
	_c->line_to(_r,_t);
	_c->line_to(_r,_b);
	_c->line_to(_l,_b);
	_c->stroke(_color,_brush);
	_c->pop_transform();
}

//
//
//

GrowingRect::GrowingRect(Coord l,Coord b,Coord r,Coord t,Canvas *c): SlidingRect(0,0,1,1,c)
{
	Transformer *tf=new Transformer(r-l,0,0,t-b,l,b);

	transformer(tf);
	tf->unref();
}

GrowingRect::~GrowingRect()
{
}

void GrowingRect::flex_point(Coord x,Coord y)
{
	Coord x1,y1,x2,y2;
	Transformer *t;

	getrect(x1,y1,x2,y2);
	t=new Transformer(x-x1,0,0,y-y1,x1,y1);
	transformer(t);
	t->unref();
}
