/*
Copyright (c) 1993, Clemens Wagner
 
Permission to use, copy, modify, and distribute this software and 
its documentation for any purpose is hereby granted without fee, provided
that
(i) you have alays to distribute the whole archive
(ii) the above copyright notices and this permission notice appear in
all copies of the software and related documentation, changed or new
code has to be marked
(iii) else, there are the same conditions as for InterViews

THE SOFTWARE IS PROVIDED "AS-IS" AND WITHOUT WARRANTY OF ANY KIND, 
EXPRESS, IMPLIED OR OTHERWISE, INCLUDING WITHOUT LIMITATION, ANY 
WARRANTY OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.  
 */

#include <Graphic/graphicmaster.h>
#include <Graphic/glyphviewer.h>
#include <Graphic/tool.h>
#include <Graphic/tool_button.h>

#include <IV-look/kit.h>

#include <InterViews/canvas.h>
#include <InterViews/color.h>
#include <InterViews/geometry.h>
#include <InterViews/telltale.h>
#include <InterViews/transformer.h>

//
//
//

Button *tool_button(Glyph *g,TelltaleFlags f,TelltaleGroup *tg,Action *a)
{
	WidgetKit &kit=*WidgetKit::instance();
	Button *b;
	PushButtonState *pbs=new PushButtonState(f,a);

	kit.begin_style("PushButton","Button");
	b=new Button(kit.push_button_look(g,pbs),kit.style(),pbs,nil);
	pbs->join(tg);
	kit.end_style();
	return b;
}

//
//
//

PushButtonState::PushButtonState(const TelltaleFlags f,Action *a): TelltaleState (f)
{
	_action=a;
}

void PushButtonState::set(const TelltaleFlags f,boolean flag)
{
	if(flag && (f==is_active))
	{
		TelltaleState::set(is_active|is_chosen,flag);
		if(_action!=nil)
			_action->execute();
	}
	else
		if(!flag && f==is_chosen)
			TelltaleState::set(is_active|is_chosen,flag);
		else
			if(!flag && f==is_active)
			{
			}
			else
				if(flag && f==is_chosen)
				{
				}
				else
					TelltaleState::set(f,flag);
}

//
//
//

ToolAction::ToolAction(Tool *t)
{
	_tool=t;
	Resource::ref(_tool);
}

ToolAction::~ToolAction()
{
	Resource::unref(_tool);
}

void ToolAction::execute()
{
	_tool->graphic_master()->glyphviewer()->tool(_tool);
}

//
//
//

GraphicGlyph::GraphicGlyph(Graphic *gr,const Color *active): MonoGlyph(gr)
{
	Resource::ref(_active=active);
	_state=nil;
}

GraphicGlyph::~GraphicGlyph()
{
}

void GraphicGlyph::request(Requisition &rq) const
{
	MonoGlyph::request(rq);
}

void GraphicGlyph::allocate(Canvas *c,const Allocation &a,Extension &ext)
{
	MonoGlyph::allocate(c,a,ext);
	//ext.merge(c,a);
}

void GraphicGlyph::draw(Canvas *c,const Allocation &a) const
{
	Graphic *g=(Graphic *) body();
	const Color *col;

	if(_state!=nil && _state->test(TelltaleState::is_chosen))
	{
		Resource::ref(col=g->fill());
		g->fill(_active);
	}
	g->translate(a.x(),a.y());
	MonoGlyph::draw(c,a);
	g->translate(-a.x(),-a.y());
	if(_state!=nil && _state->test(TelltaleState::is_chosen))
	{
		g->fill(col);
	}
}

void GraphicGlyph::undraw()
{
	MonoGlyph::undraw();
}

void GraphicGlyph::state(TelltaleState *s)
{
	Resource::ref(s);
	Resource::unref(_state);
	_state=s;
}

TelltaleState *GraphicGlyph::state() const
{
	return _state;
}
